/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef __TIZEN_CAPI_NETWORK_USER_AWARENESS_PRIVATE_H__
#define __TIZEN_CAPI_NETWORK_USER_AWARENESS_PRIVATE_H__

#include <user-awareness-log.h>
#include <glib.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Returns if expr is true.
 * @since_tizen 5.5
 */
#define ret_if(expr) \
	do { \
		if (expr) { \
			UA_DBG("(%s) return", #expr); \
			return; \
		} \
	} while (0)

/**
 * @brief Returns val if expr is true.
 * @since_tizen 5.5
 */
#define retv_if(expr, val) \
	do { \
		if (expr) { \
			UA_DBG("(%s) return", #expr); \
			return (val); \
		} \
	} while (0)

/**
 * @brief Returns error code if expr is NULL.
 * @since_tizen 5.5
 */
#define UA_VALIDATE_INPUT_PARAMETER(arg) \
	if (arg == NULL) { \
		UA_ERR("INVALID_PARAMETER(%s is NULL)", #arg); \
		return UA_ERROR_INVALID_PARAMETER; \
	}

/**
 * @brief Prints arg.
 * @since_tizen 5.5
 */
#define UA_PRINT_DEVICE_HANDLE(arg) \
		UA_INFO("Device Handle [%p]", #arg); \

/**
 * @brief Prints arg.
 * @since_tizen 5.5
 */
#define UA_PRINT_USER_HANDLE(arg) \
		UA_INFO("User Handle [%p]", #arg); \

/**
 * @brief Checks if h1 is not in the list.
 * @since_tizen 5.5
 */
#define UA_VALIDATE_HANDLE(h1, list) \
	{ \
		GSList *l; \
		bool valid = FALSE; \
		for (l = list; l; l = g_slist_next(l)) { \
			void *h2 = (void *)l->data; \
			if (h1 == h2) { \
				UA_INFO("Handle matched [%p]", h2); \
				valid = TRUE; break; \
			} \
		} \
		if (valid == FALSE) { \
			UA_ERR("Handle [%p] did not match with any stored handles!!", h1); \
			return UA_ERROR_INVALID_PARAMETER; \
		} \
	} \

/**
 * @brief Returns true if h1 is in the list, else return false.
 * @since_tizen 5.5
 */
#define UA_EXIST_HANDLE(h1, list) \
	{ \
		GSList *l; \
		bool valid = FALSE; \
		for (l = list; l; l = g_slist_next(l)) { \
			void *h2 = (void *)l->data; \
			if (h1 == h2) { \
				UA_INFO("Handle matched [%p]", h2); \
				valid = TRUE; break; \
			} \
		} \
		if (valid == TRUE) { \
			UA_ERR("Handle [%p] match with any stored handles!!", h1); \
			return UA_ERROR_ALREADY_DONE; \
		} \
	} \

#define UA_ALL_SENSOR_PRESENCE_DETECTION_WAIT_TIME 65 /**< All sensor presence detection wait time */

/**
 * @brief Minimum detection window size.
 * @since_tizen 5.5
 */
#define UA_MAX_DETECTION_WINDOW 290 /**< 290 seconds */

/**
 * @brief Minimum detection cycle.
 * @since_tizen 5.5
 */
#define UA_MIN_DETECTION_CYCLE 300 /**< 5 min */

/**
 * @brief Default service name.
 * @since_tizen 5.5
 */
#define UA_SERVICE_DEFAULT "ua.service.default" /**< Default service name */

/**
 * @brief Event type for user event callback data structure.
 * @since_tizen 5.5
 */
typedef enum {
	UA_USER_EVENT_DEVICE_ADDED = 0, /**< Device added for a user */
	UA_USER_EVENT_MAX /**< Event Max. */
} ua_user_event_e;

/**
 * @brief Callback data structure.
 * @since_tizen 5.5
 */
typedef struct {
	const void *callback; /**< Callback function pointer */
	void *user_data; /**< User data */
} ua_callback_s;

/**
 * @brief Monitor data structure.
 * @since_tizen 5.5
 */
typedef struct {
	unsigned int sensor_bitmask; /**< Bit-mask for sensors which be supported */
	unsigned int presence_detected_bitmask; /**< Detetected presence bitmask */
	unsigned int absence_detected_bitmask; /**< Detected Absence bitmask */

	unsigned int presence_bitmask_and; /**< AND operation bitmask for presence */
	unsigned int presence_bitmask_or; /**< OR operation bitmask for presence */
	unsigned int absence_bitmask_and; /**< AND operation bitmask for absence */
	unsigned int absence_bitmask_or; /**< OR operation bitmask for absence */

	ua_detection_mode_e presence_mode; /**< Presence mode */
	ua_detection_mode_e absence_mode; /**< Absence mode */
	gboolean presence_detection_started; /**< Is presence detection started */
	gboolean absence_detection_started; /**< Is absence detection started */
	gboolean scan_device_started; /**< Is scan device started */
	gboolean internal_presence_started; /**< Is Internally scan device started */
	unsigned int presence_detection_timer; /**< Presence detection timer */
	unsigned int absence_detection_timer; /**< Absence detection timer */
	ua_callback_s sensor_state_cb; /**< Callback to let apps know sensors added or not */
	ua_callback_s absence_user_cb; /**< Callback to let apps know user is absence */
	ua_callback_s presence_user_cb; /**< Callback to let apps know user is presence */
	ua_scan_completed_cb scan_device_cb; /**< Callback to let apps know scanned registered devices */
	ua_presence_detected_cb presence_cb; /**< User presence detection callback */
	ua_absence_detected_cb absence_cb; /**< Absence detection callback */
	char *service; /**< Service name */
	GSList *user_state; /**< User current state i.e presence or absence */
	void *user_data; /**< User data */
} ua_monitor_s;

/**
 * @brief User state data structure.
 * @since_tizen 5.5
 */
typedef struct {
	char *account; /**< Account */
	unsigned int sensor_bitmask; /**< Detected sensor type bit mask */
} ua_user_state_info_s;

/**
 * @brief Sensor state data structure.
 * @since_tizen 5.5
 */
typedef struct {
	ua_monitor_s *monitor; /**< Monitor handle */
	ua_sensor_e type; /**< Bitmask to express the supported sensors */
	unsigned short absence_period; /**< Absence period */
	unsigned int absence_timer; /**< Absence timer */
	unsigned int detection_timer; /**< Detection timer */
} ua_sensors_s;

/**
 * @brief User info. data structure.
 * @since_tizen 5.5
 */
typedef struct {
	ua_user_h user_handle; /**< User handle */
	ua_presence_state_e state; /**< Presence state */
	unsigned int sensor_bitmask; /**< Detected user devices bitmask */
	char *account; /**< Account */
	char *name; /**< User name */
	gboolean isadded; /**< Is the user addition completed? */
	gboolean default_user; /**< Is it a default user? */
	gboolean create_by_app; /**< Did app add this user information? */
	long last_presence_timestamp; /**< timestamp of last presence */
} ua_user_info_s;

/**
 * @brief Service info. data structure.
 * @since_tizen 5.5
 */
typedef struct {
	ua_service_h service_handle; /**< Service handle */
	char *name; /**< Service name */
	gboolean isadded; /**< Is the service addition completed? */
	gboolean default_service; /**< Is it a default service? */
	gboolean create_by_app; /**< Did app add this service information? */
} ua_service_info_s;

/**
 * @brief Detection type.
 * @since_tizen 5.5
 */
typedef enum {
	UA_PRESENCE_DETECTION = 0x00, /**< Presence detcection */
	UA_ABSENCE_DETECTION /**< Absence detection */
} ua_detection_type_e;

/**
 * @brief BLE payload data structure, used for filter.
 * @since_tizen 5.5
 */
typedef struct {
	char service_id; /**< service id for manufacturer data */
	char device_icon; /**< device icon for manufacturer data */
	char purpose; /**< purpose for manufacturer data */
	char *duid; /**< duid for manufacturer data */
	char *bt_mac; /**< bt mac address of device */
} ua_ble_payload_s;

/**
 * @brief Device info data structure.
 * @since_tizen 5.5
 */
typedef struct {
	ua_user_h user; /**< User handle */
	ua_device_h handle; /**< Device handle */
	char *mac; /**< MAC address */
	char *bssid; /**< BSSID */
	char *ipv4; /**< IPv4 address */
	char *ipv6; /**< IPv6 address */
	char *device_id; /**< Mobile id (e.g. BT MAC or UUID) */
	ua_mac_type_e type; /**< Connectivity technology type using this MAC address */
	bool pairing_required; /**< Does it need pairing? */
	ua_presence_state_e state; /**< Presence state */
	ua_os_type_e os; /**< OS type */
	gboolean isadded; /**< Is the device addition completed? */
	gboolean create_by_app; /**< Did app add this device information? */
	gboolean discriminant; /**< Determines whether to judge PRESENCE/ABSENCE */
	long last_presence_timestamp; /**< timestamp of last presence */
	ua_ble_payload_s *payload; /**< BLE payload to filter device*/
} ua_dev_info_s;

/**
 * @brief Calls the scan callback when the scan is complete.
 * @since_tizen 5.5
 *
 * @param[in] result Result code
 * @param[in] uam_info Device information
 *
 * @exception
 * @pre
 * @post
 */
void _ua_monitor_handle_scanned_device(int result, uam_device_info_s *uam_info);

/**
 * @brief Clears data related to scan in the monitor.
 * @since_tizen 5.5
 *
 * @param[in] result Result code.
 *
 * @exception
 * @pre
 * @post
 */
void _ua_monitor_handle_scan_complete(int result);

/**
 * @brief Calls the presence callback when detected registerd devices.
 * @since_tizen 5.5
 *
 * @param[in] bitmask Bitmask for sensors.
 * @param[in] service Service information.
 * @param[in] account Account information.
 *
 * @exception
 * @pre
 * @post
 */
void _ua_monitor_handle_user_presence_detected(unsigned int bitmask,
					       char *service, char *account,
					       long timestamp, char *device_id);

/**
 * @brief Stops monitoring detection.
 * @since_tizen 5.5
 *
 * @exception
 * @pre
 * @post
 */
void _ua_monitor_handle_detection_stopped(void);

/**
 * @brief Changes sensor state.
 * @since_tizen 5.5
 *
 * @param[in] bitmask Bitmask for sensors.
 * @param[in] ready Ready or not.
 *
 * @exception
 * @pre
 * @post
 */
void _ua_monitor_handle_sensor_state(unsigned int bitmask, gboolean ready);

/**
 * @brief Calls the absence callback when detected registerd devices.
 * @since_tizen 5.5
 *
 * @param[in] bitmask Bitmask for sensors
 * @param[in] service Service information
 * @param[in] account Account information
 *
 * @exception
 * @pre
 * @post
 */
void _ua_monitor_handle_user_absence_detected(
		unsigned int bitmask, char *service, char *account);

/**
 * @brief Destroys a monitor handle.
 * @since_tizen 5.5
 *
 * @param[in] data Monitor handle.
 *
 * @exception
 * @pre
 * @post
 */
void _ua_free_ua_monitor_t(gpointer data);

/**
 * @brief Destroys a user handle.
 * @since_tizen 5.5
 *
 * @param[in] user User handle.
 *
 * @exception
 * @pre
 * @post
 */
void _ua_free_ua_user_info_t(gpointer data);

/**
 * @brief Destroys a service handle.
 * @since_tizen 5.5
 *
 * @param[in] service Service handle.
 *
 * @exception
 * @pre
 * @post
 */
void _ua_free_ua_service_info_s(gpointer data);

/**
 * @brief Create a payload handle.
 * @since_tizen 5.5
 *
 * @param[out] payload_handle Payload handle.
 *
 * @exception
 * @pre
 * @post
 *
 * @see _ua_free_ua_ble_payload_s()
 */
void _ua_create_ua_ble_payload_s(ua_ble_payload_s **payload_handle);

/**
 * @brief Destroys a payload handle.
 * @since_tizen 5.5
 *
 * @param[in] payload Payload handle.
 *
 * @exception
 * @pre
 * @post
 *
 * @see _ua_create_ua_ble_payload_s()
 */
void _ua_free_ua_ble_payload_s(gpointer data);

/**
 * @brief Return payload handle from uam_ble_payload_s.
 * @since_tizen 5.5
 *
 * @param[in] uam_payload Payload handle.
 *
 * @return Payload handle
 * @exception
 * @pre
 * @post
 *
 * @see _ua_free_ua_ble_payload_s()
 */
ua_ble_payload_s* _ua_get_payload_from_uam(uam_ble_payload_s *uam_payload);

/**
 * @brief Creates payload handle of uam_ble_payload_s type from ua_ble_payload_s.
 * @since_tizen 5.5
 *
 * @param[in] payload Payload handle.
 * @param[out] uam_payload Payload handle.
 *
 * @exception
 * @pre
 * @post
 */
void _ua_get_uam_payload_from_ua(uam_ble_payload_s *uam_payload, ua_ble_payload_s *payload);

/**
 * @brief Returns a cloned copy of payload handle.
 * @since_tizen 5.5
 *
 * @param[in] org_payload Payload handle.
 *
 * @return Payload handle
 * @exception
 * @pre
 * @post
 */
ua_ble_payload_s* _ua_payload_clone(ua_ble_payload_s *org_payload);


/**
 * @brief Destroys a device handle.
 * @since_tizen 5.5
 *
 * @param[in] data Device handle.
 *
 * @exception
 * @pre
 * @post
 */
void _ua_free_ua_device_info_t(gpointer data);

/**
 * @brief Gets a device handle.
 * @since_tizen 5.5
 *
 * @remarks The returned value should not be released.
 * @remarks The returned value is managed by the platform and will be released when destrying user list.
 *
 * @param[in] uam_info Device info. handle
 *
 * @return not null on success, otherwise error
 *
 * @exception
 * @pre
 * @post
 */
ua_dev_info_s* _ua_get_device_info_from_uam(uam_device_info_s *uam_info);

/**
 * @brief Sets data in user info. state callback.
 * @since_tizen 5.5
 *
 * @param[in] account Account.
 * @param[in] state #UA_PRESENCE_STATE_ACTIVE or #UA_PRESECE_STATE_INACTIVE.
 * @param[in] sensor_bitmask Sensor bitmask.
 *
 * @return 0 on success, otherwise a negative error value
 * @retval #UA_ERROR_NONE Successful
 * @retval #UA_ERROR_NO_DATA No data
 *
 * @exception
 * @pre
 * @post
 */
int _ua_mark_user_info_state_db(char *account, int state, unsigned sensor_bitmask);

/**
 * @brief Checks whether user's detection type.
 * @since_tizen 5.5
 *
 * @param[in] bitmask Sensor bitmask.
 *
 * @return true on success, otherwise error
 * @retval true Successful
 * @retval false Failed
 *
 * @exception
 * @pre
 * @post
 */
bool _ua_check_all_users_absence_any(unsigned int bitmask);

/**
 * @brief Adds a user info to the list.
 * @since_tizen 5.5
 *
 * @param[in] ua_info User info.
 *
 * @return 0 on success, otherwise a negative error value
 * @retval #UA_ERROR_NONE Successful
 * @retval #UA_ERROR_OUT_OF_MEMORY Out of memory
 *
 * @exception
 * @pre
 * @post
 */
int _ua_user_add_info_to_list(ua_user_info_s* ua_info);

/**
 * @brief Adds a user info to the list in ua-manager.
 * @since_tizen 5.5
 *
 * @param[in] uam_info User info.
 *
 * @return 0 on success, otherwise a negative error value
 * @retval #UA_ERROR_NONE Successful
 * @retval #UA_ERROR_OUT_OF_MEMORY Out of memory
 *
 * @exception
 * @pre
 * @post
 */
int _ua_user_add_info_to_list_from_uapi_data(uam_user_info_s *uam_info);

/**
 * @brief Removes a user info to the list.
 * @since_tizen 5.5
 *
 * @param[in] account User account.
 *
 * @return 0 on success, otherwise a negative error value
 * @retval #UA_ERROR_NONE Successful
 * @retval #UA_ERROR_NO_DATA No data
 *
 * @exception
 * @pre
 * @post
 */
int _ua_remove_user_info_from_list(char *account);

/**
 * @brief Adds a service info to the list.
 * @since_tizen 5.5
 *
 * @param[in] ua_info Service info.
 *
 * @return 0 on success, otherwise a negative error value
 * @retval #UA_ERROR_NONE Successful
 * @retval #UA_ERROR_OUT_OF_MEMORY Out of memory
 *
 * @exception
 * @pre
 * @post
 */
int _ua_service_add_info_to_list(ua_service_info_s* ua_info);

/**
 * @brief Adds a service info of initial type as uam_service_info_s to the list.
 * @since_tizen 5.5
 *
 * @param[in] uam_info Service info.
 *
 * @return 0 on success, otherwise a negative error value
 * @retval #UA_ERROR_NONE Successful
 * @retval #UA_ERROR_OUT_OF_MEMORY Out of memory
 *
 * @exception
 * @pre
 * @post
 */
int _ua_service_add_info_to_list_from_uam_data(uam_service_info_s *uam_info);

/**
 * @brief Removes a service info from the list.
 * @since_tizen 5.5
 *
 * @param[in] uam_info Service info.
 *
 * @return 0 on success, otherwise a negative error value
 * @retval #UA_ERROR_NONE Successful
 * @retval #UA_ERROR_NO_DATA No data
 *
 * @exception
 * @pre
 * @post
 */
int _ua_service_remove_info_from_list(uam_service_info_s *uam_info);

/**
 * @brief Adds a device info. to DB in ua-manager.
 * @since_tizen 5.5
 *
 * @param[in] result Result code.
 * @param[in] uam_info Device information.
 *
 * @exception
 * @pre
 * @post
 */
void _ua_handle_device_added(int result, uam_device_info_s *uam_info);

/**
 * @brief Removes a device info. from DB in ua-manager.
 * @since_tizen 5.5
 *
 * @param[in] result Result code.
 * @param[in] uam_info Device information.
 *
 * @exception
 * @pre
 * @post
 */
void _ua_handle_device_removed(int result, uam_device_info_s *uam_info);

/**
 * @brief Callback to get the service handle.
 * @since_tizen 5.5
 *
 * @remarks The @a service_handle should not be released.
 * @remarks The @a service_handle can be used only in the callback.
 *
 * @param[in] service_handle The service handle.
 * @param[in] user_data User data passed in _ua_intr_foreach_registered_services().
 *
 * @exception
 * @pre
 * @post
 *
 * @see _ua_intr_foreach_registered_services()
 */
typedef bool (*_ua_intr_registered_service_cb)(ua_service_h service_handle, void *user_data);

/**
 * @brief Retrieves the service handle of all the registered services in ua-manager.
 * @since_tizen 5.5
 *
 * @param[in] foreach_cb Callback function to be invoked with service handle.
 * @param[in] user_data The user data to be passed when callback is called.
 *
 * @return 0 on success, otherwise a negative error value
 * @retval #UA_ERROR_NONE Successful
 * @retval #UA_ERROR_INVALID_PARAMETER Invalid parameter
 *
 * @exception
 * @pre
 * @post
 *
 * @see _ua_intr_registered_service_cb()
 */
int _ua_intr_foreach_registered_services(_ua_intr_registered_service_cb foreach_cb,
		void *user_data);

/**
 * @brief Retrieves the service handle of all the registered services.
 * @since_tizen 5.5
 *
 * @param[in] foreach_cb Callback function to be invoked with service handle.
 * @param[in] user_data The user data to be passed when callback is called.
 *
 * @return 0 on success, otherwise a negative error value
 * @retval #UA_ERROR_NONE Successful
 * @retval #UA_ERROR_INVALID_PARAMETER Invalid parameter
 *
 * @exception
 * @pre
 * @post
 *
 * @see ua_service_added_service_cb()
 */
int _ua_foreach_registered_services(ua_service_added_service_cb foreach_cb,
		void *user_data);

/**
 * @brief Gets default service data.
 * @since_tizen 5.5
 *
 * @return 0 on success, otherwise a negative error value
 * @retval #UA_ERROR_NONE Successful
 * @retval #UA_ERROR_OUT_OF_MEMORY Out of memory
 *
 * @exception
 * @pre
 * @post
 */
int _ua_intr_get_default_service(void);

/**
 * @brief Callback to get the user handle added to the service.
 * @since_tizen 5.5
 *
 * @remarks The @a user_handle should not be released.
 * @remarks The @a user_handle can be used only in the callback.
 *
 * @param[in] user_handle The user handle.
 * @param[in] user_data User data passed in _ua_intr_foreach_registered_users().
 *
 * @exception
 * @pre
 * @post
 */
typedef bool (*_ua_intr_registered_user_cb)(ua_user_h user_handle, void *user_data);

/**
 * @brief Retrieves the user handle of all the registered users in ua-manager.
 * @since_tizen 5.5
 *
 * @param[in] foreach_cb Callback function to be invoked with user handle.
 * @param[in] user_data The user data to be passed when callback is called.
 *
 * @return 0 on success, otherwise a negative error value
 * @retval #UA_ERROR_NONE Successful
 * @retval #UA_ERROR_INVALID_PARAMETER Invalid parameter
 *
 * @exception
 * @pre
 * @post
 *
 * @see _ua_intr_registered_user_cb()
 */
int _ua_intr_foreach_registered_users(_ua_intr_registered_user_cb foreach_cb,
		void *user_data);

/**
 * @brief Retrieves the user handle of all the registered users.
 * @since_tizen 5.5
 *
 * @param[in] foreach_cb Callback function to be invoked with user handle.
 * @param[in] user_data The user data to be passed when callback is called.
 *
 * @return 0 on success, otherwise a negative error value
 * @retval #UA_ERROR_NONE Successful
 * @retval #UA_ERROR_INVALID_PARAMETER Invalid parameter
 *
 * @exception
 * @pre
 * @post
 *
 * @see ua_registered_user_cb()
 */
int _ua_foreach_registered_users(ua_registered_user_cb foreach_cb,
		void *user_data);

/**
 * @brief Gets default user data.
 * @since_tizen 5.5
 *
 * @return 0 on success, otherwise a negative error value
 * @retval #UA_ERROR_NONE Successful
 * @retval #UA_ERROR_OUT_OF_MEMORY Out of memory
 *
 * @exception
 * @pre
 * @post
 */
int _ua_intr_get_default_user(void);

/**
 * @brief Gets a user handle by the account.
 * @since_tizen 5.5
 *
 * @remarks The returned value should not be released.
 * @remarks The returned value is managed by the platform and will be released when destrying user list.
 *
 * @param[in] account Account
 *
 * @return not null on success, otherwise error
 *
 * @exception
 * @pre
 * @post
 */
ua_user_h _ua_get_user_handle_by_account(const char *account);

/**
 * @brief set last presence timestamp in user handle.
 * @since_tizen 5.5
 *
 * @param[in] user_handle User handle.
 * @param[in] timestamp The timestamp to be updated in user handle.
 *
 * @return not null on success, otherwise error
 *
 * @exception
 * @pre
 * @post
 */
void _ua_set_user_last_presence_timestamp(ua_user_h user_handle,
					  long timestamp);

/**
 * @brief Checks whether a device exists in device list.
 * @since_tizen 5.5
 *
 * @param[in] device_id The device id.
 * @param[in] mac The device MAC address.
 * @param[in] type The device type.
 * @param[in] is_exist Whether exists or not.
 *
 * @return 0 on success, otherwise a negative error value
 * @retval #UA_ERROR_NONE Successful
 * @retval #UA_ERROR_INVALID_PARAMETER Invalid parameter
 *
 * @exception
 * @pre
 * @post
 */
int _ua_is_device_exist(char *device_id, char *mac, ua_mac_type_e type,
		gboolean *is_exist);

/**
 * @brief Gets the service list.
 * @since_tizen 5.5
 *
 * @remarks The returned value should not be released.
 * @remarks The returned value is managed by the platform and will be released when destroying service list.
 *
 * @return not null on success, otherwise error
 *
 * @exception
 * @pre
 * @post
 */
GSList *_ua_service_get_services(void);

/**
 * @brief Gets the user list.
 * @since_tizen 5.5
 *
 * @remarks The returned value should not be released.
 * @remarks The returned value is managed by the platform and will be released when destroying user list.
 *
 * @return not null on success, otherwise error
 *
 * @exception
 * @pre
 * @post
 */
GSList *_ua_user_get_users(void);

#ifdef __cplusplus
}
#endif
#endif /* __TIZEN_CAPI_NETWORK_USER_AWARENESS_PRIVATE_H__ */
