/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <glib.h>
#include <ua-api.h>
#include <user-awareness.h>
#include <user-awareness-log.h>
#include <user-awareness-private.h>
#include <user-awareness-util.h>

GSList *ua_users_list;

extern  GSList *ua_devices_list;
extern ua_callback_s user_callbacks[UA_USER_EVENT_MAX];

static ua_user_info_s *__ua_get_user_from_list(const char *account)
{
	GSList *l;
	ua_user_info_s *user_info;

	retv_if(NULL == account, NULL);

	for (l = ua_users_list; l; l = g_slist_next(l)) {
		user_info = (ua_user_info_s *)l->data;

		if (!g_strcmp0(account, user_info->account))
			return user_info;
	}

	return NULL;
}

void _ua_free_ua_user_info_t(gpointer data)
{
	FUNC_ENTRY;
	ua_user_info_s *user = data;

	ret_if(NULL == user);

	if (user->account)
		g_free(user->account);

	if (user->name)
		g_free(user->name);

	g_free(user);

	FUNC_EXIT;
}

/* LCOV_EXCL_START */
int _ua_mark_user_info_state_db(char *account, int state,
	unsigned sensor_bitmask)
{
	FUNC_ENTRY;
	int ret = UA_ERROR_NO_DATA;
	GSList *l;
	ua_user_info_s *u;
	UA_VALIDATE_INPUT_PARAMETER(account);

	for (l = ua_users_list; l; l = g_slist_next(l)) {
		u = (ua_user_info_s *)l->data;

		if (!g_strcmp0(account, u->account)) {
			UA_INFO("User found [%s]", u->account);
			if (state == UA_PRESENCE_STATE_ACTIVATE) {
				u->sensor_bitmask |= sensor_bitmask;
				u->state = state;
			} else if (state == UA_PRESENCE_STATE_INACTIVATE) {
				u->sensor_bitmask &= ~sensor_bitmask;
				if (u->sensor_bitmask == 0)
					u->state = state;
			} else
				UA_INFO("Invalid state");

			ret = UA_ERROR_NONE;
			break;
		}
	}

	FUNC_EXIT;
	return ret;
}

bool _ua_check_all_users_absence_any(unsigned int bitmask)
{
	FUNC_ENTRY;
	GSList *l;
	ua_user_info_s *u;

	for (l = ua_users_list; l; l = g_slist_next(l)) {
		u = (ua_user_info_s *)l->data;

		if (u->sensor_bitmask & bitmask) {
			UA_INFO("User found [%s]", u->account);
			return false;
		}
	}

	return true;
}
/* LCOV_EXCL_STOP */

int _ua_user_add_info_to_list_from_uapi_data(uam_user_info_s *uam_info)
{
	FUNC_ENTRY;
	ua_user_info_s* user_info;
	UA_VALIDATE_INPUT_PARAMETER(uam_info);

	if (NULL != __ua_get_user_from_list(uam_info->account)) {
			UA_INFO("User already in list(added on user_create) [%s]", uam_info->account);
			return UA_ERROR_NONE;
	}

	user_info = g_malloc0(sizeof(ua_user_info_s));

	if (!user_info) {
		UA_ERR("g_malloc0 failed");
		return UA_ERROR_OUT_OF_MEMORY;
	}

	user_info->account = g_strdup(uam_info->account);

	if (user_info->account == NULL) {
		UA_ERR("g_malloc0 failed");
		_ua_free_ua_user_info_t(user_info);
		return UA_ERROR_OUT_OF_MEMORY;
	}

	user_info->name = g_strdup(uam_info->name);

	if (user_info->name == NULL) {
		UA_ERR("g_malloc0 failed");
		_ua_free_ua_user_info_t(user_info);
		return UA_ERROR_OUT_OF_MEMORY;
	}

	user_info->isadded = true;
	user_info->user_handle = (ua_user_h)user_info;

	/* Set user state by default INACTIVATE as this field will be use
	 * during absence detection where by default all users are in INACTIVATE state
	 * user will go in ACTIVATE state after presence is detected for thst user.
	 */
	user_info->state = UA_PRESENCE_STATE_INACTIVATE;
	user_info->sensor_bitmask = 0;
	ua_users_list = g_slist_append(ua_users_list, user_info);

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int _ua_user_add_info_to_list(ua_user_info_s* ua_info)
{
	FUNC_ENTRY;
	ua_user_info_s* user_info;
	UA_VALIDATE_INPUT_PARAMETER(ua_info);

	if (NULL != __ua_get_user_from_list(ua_info->account))
			return UA_ERROR_ALREADY_DONE; // LCOV_EXCL_LINE

	user_info = g_malloc0(sizeof(ua_user_info_s));

	if (!user_info) {
		/* LCOV_EXCL_START */
		UA_ERR("g_malloc0 failed");
		return UA_ERROR_OUT_OF_MEMORY;
		/* LCOV_EXCL_STOP */
	}

	user_info->account = g_strdup(ua_info->account);

	if (user_info->account == NULL) {
		/* LCOV_EXCL_START */
		UA_ERR("g_malloc0 failed");
		_ua_free_ua_user_info_t(user_info);
		return UA_ERROR_OUT_OF_MEMORY;
		/* LCOV_EXCL_STOP */
	}

	user_info->name = g_strdup(ua_info->name);

	if (user_info->name == NULL) {
		/* LCOV_EXCL_START */
		UA_ERR("g_malloc0 failed");
		_ua_free_ua_user_info_t(user_info);
		return UA_ERROR_OUT_OF_MEMORY;
		/* LCOV_EXCL_STOP */
	}

	user_info->isadded = true;
	user_info->user_handle = (ua_user_h)user_info;

	/* Set user state by default INACTIVATE as this field will be use
	 * during absence detection where by default all users are in INACTIVATE state
	 * user will go in ACTIVATE state after presence is detected for thst user.
	 */
	user_info->state = UA_PRESENCE_STATE_INACTIVATE;
	user_info->sensor_bitmask = 0;
	ua_users_list = g_slist_append(ua_users_list, user_info);

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int _ua_remove_user_info_from_list(char *account)
{
	FUNC_ENTRY;
	int ret = UA_ERROR_NO_DATA;
	GSList *l;
	GSList *m;
	ua_user_info_s *u;
	ua_dev_info_s *dev;
	UA_VALIDATE_INPUT_PARAMETER(account);


	for (l = ua_users_list; l; l = g_slist_next(l)) {
		u = (ua_user_info_s *)l->data;

		if (!g_strcmp0(account, u->account)) {
			if (u->create_by_app == false) {
				UA_INFO("User found [%s]", u->account);

				for (m = ua_devices_list; m; m = g_slist_next(m)) {
					dev = (ua_dev_info_s *)m->data;
					if (u == dev->user)
						dev->user = NULL;
				}

				ua_users_list = g_slist_remove(ua_users_list, u);
				_ua_free_ua_user_info_t(u);
			} else {
				UA_INFO("User created by APP");
				u->isadded = false;
			}
			ret = UA_ERROR_NONE;
			break;
		}
	}

	FUNC_EXIT;
	return ret;
}

ua_user_h _ua_get_user_handle_by_account(const char *account)
{
	FUNC_ENTRY;
	GSList *l;
	ua_user_info_s *u;

	if (account == NULL) {
		UA_INFO("Account is NULL");
		return NULL;
	}

	for (l = ua_users_list; l; l = g_slist_next(l)) {
		u = (ua_user_info_s *)l->data;

		if (!g_strcmp0(account, u->account)) {
			UA_INFO("User found [%s]", u->account);
			return (u->user_handle);
		}
	}

	FUNC_EXIT;
	return NULL;
}

void _ua_set_user_last_presence_timestamp(ua_user_h user_handle,
	long timestamp)
{
	FUNC_ENTRY;
	ua_user_info_s *user = (ua_user_info_s *)user_handle;

	if (user == NULL) {
		UA_INFO("user_handle is NULL");
		FUNC_EXIT;
		return;
	}

	user->last_presence_timestamp = timestamp;

	FUNC_EXIT;
}

 int _ua_intr_foreach_registered_users(_ua_intr_registered_user_cb foreach_cb,
	void *user_data)
{
	int ret;
	UA_VALIDATE_INPUT_PARAMETER(foreach_cb);

	ret = _ua_foreach_registered_users(foreach_cb, user_data);
	return ret;
}

int _ua_intr_get_default_user(void)
{
	FUNC_ENTRY;
	int ret;
	uam_user_info_s uam_user;
	ua_user_info_s *user = NULL;
	GSList *l;

	ret = _ua_get_error_code(_uam_get_default_user(&uam_user));
	if (UA_ERROR_NONE != ret) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		/* LCOV_EXCL_STOP */
		return ret;
	}

	for (l = ua_users_list; l; l = g_slist_next(l)) {
		user = (ua_user_info_s *)l->data;

		if (!g_strcmp0(uam_user.account, user->account)) {
			UA_INFO("User found [%s]", user->account);
			user->default_user = true;
			user->isadded = true;
			return UA_ERROR_NONE;
		}
	}

	/* LCOV_EXCL_START */
	user = g_malloc0(sizeof(ua_user_info_s));
	if (!user) {
		UA_ERR("Failed to allocate memory");
		return UA_ERROR_OUT_OF_MEMORY;
	}
	user->account = g_strdup(uam_user.account);
	user->name = g_strdup(uam_user.name);
	user->state = UA_PRESENCE_STATE_INACTIVATE;
	user->sensor_bitmask = 0;
	user->user_handle = (ua_user_h)user;
	user->default_user = true;
	user->isadded = true;

	ua_users_list = g_slist_append(ua_users_list, user);
	FUNC_EXIT;
	/* LCOV_EXCL_STOP */
	return UA_ERROR_NONE;
}

/* LCOV_EXCL_START */
GSList *_ua_user_get_users(void)
{
	FUNC_ENTRY;
	FUNC_EXIT;
	return ua_users_list;
}

int _ua_foreach_registered_users(ua_registered_user_cb foreach_cb,
	void *user_data)
{
	FUNC_ENTRY;
	int i;
	int ret;
	GPtrArray *user_list = NULL;
	uam_user_info_s *ptr;

	UA_VALIDATE_INPUT_PARAMETER(foreach_cb);

	user_list = g_ptr_array_new();
	retv_if(NULL == user_list, UA_ERROR_OUT_OF_MEMORY);

	ret = _ua_get_error_code(_uam_get_registered_users(&user_list));
	if (UA_ERROR_NONE != ret) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		g_ptr_array_free(user_list, TRUE);
		/* LCOV_EXCL_STOP */
		return ret;
	}

	for (i = 0; i < user_list->len; i++) {
		ptr = g_ptr_array_index(user_list, i);
		if (ptr) {
			ua_user_info_s* user_info;
			user_info = g_malloc0(sizeof(ua_user_info_s));

			if (!user_info) {
				/* LCOV_EXCL_START */
				UA_ERR("g_malloc0 failed");
				ret = UA_ERROR_OUT_OF_MEMORY;
				_ua_free_ua_user_info_t(user_info);
				/* LCOV_EXCL_STOP */
				goto done;
			}

			user_info->account = g_strdup(ptr->account);

			if (user_info->account == NULL) {
				/* LCOV_EXCL_START */
				UA_ERR("g_malloc0 failed");
				ret = UA_ERROR_OUT_OF_MEMORY;
				_ua_free_ua_user_info_t(user_info);
				/* LCOV_EXCL_STOP */
				goto done;
			}

			user_info->name = g_strdup(ptr->name);

			if (user_info->name == NULL) {
				/* LCOV_EXCL_START */
				UA_ERR("g_malloc0 failed");
				ret = UA_ERROR_OUT_OF_MEMORY;
				_ua_free_ua_user_info_t(user_info);
				/* LCOV_EXCL_STOP */
				goto done;
			}

			user_info->isadded = true;
			user_info->user_handle = (ua_user_h)user_info;
			if (!foreach_cb(user_info->user_handle, user_data)) {
				_ua_free_ua_user_info_t(user_info);
				break;
			} else
				_ua_free_ua_user_info_t(user_info);
		} else {
			/* LCOV_EXCL_START */
			UA_ERR("OPERATION_FAILED(0x%08x)",
					UA_ERROR_OPERATION_FAILED);
			ret = UA_ERROR_OPERATION_FAILED;
			/* LCOV_EXCL_STOP */
			goto done;
		}
	}

done:
	g_ptr_array_foreach(user_list, (GFunc)g_free, NULL);
	g_ptr_array_free(user_list, TRUE);

	FUNC_EXIT;
	return ret;
}


int ua_user_create(ua_user_h *user_handle)
{
	FUNC_ENTRY;
	ua_user_info_s *user = NULL;

	UA_VALIDATE_INPUT_PARAMETER(user_handle);

	user = g_malloc0(sizeof(ua_user_info_s));
	if (!user) {
		/* LCOV_EXCL_START */
		UA_ERR("g_malloc0 failed");
		return UA_ERROR_OUT_OF_MEMORY;
		/* LCOV_EXCL_STOP */
	}
	user->state = UA_PRSENCE_STATE_INVALID;
	user->account = NULL;
	user->last_presence_timestamp = 0;

	/* Add user to list of users*/
	*user_handle = (ua_user_h)user;
	user->user_handle = (ua_user_h)user;
	user->create_by_app = true;

	ua_users_list = g_slist_append(ua_users_list, user);

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_add(ua_user_h user_handle)
{
	FUNC_ENTRY;
	int ret;
	ua_user_info_s *user = (ua_user_info_s *)user_handle;
	uam_user_info_s uam_user;

	UA_VALIDATE_INPUT_PARAMETER(user_handle);
	UA_VALIDATE_HANDLE(user_handle, ua_users_list);

	retv_if(NULL == user->account, UA_ERROR_INVALID_PARAMETER);
//TODO lk, check user->is_added before proceed

	memset(&uam_user, 0, sizeof(uam_user_info_s));
	g_strlcpy(uam_user.account, user->account, UAM_USER_ACCOUNT_MAX_STRING_LEN);

	if (user->name)
		g_strlcpy(uam_user.name, user->name, UAM_USER_NAME_MAX_STRING_LEN);

	ret = _ua_get_error_code(_uam_add_user(&uam_user));
	if (UA_ERROR_NONE != ret) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		return ret;
		/* LCOV_EXCL_STOP */
	}

	user->isadded = true;
	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_remove(ua_user_h user_handle)
{
	FUNC_ENTRY;
	int ret;
	ua_user_info_s *user = (ua_user_info_s *)user_handle;
	char account[UAM_USER_ACCOUNT_MAX_STRING_LEN];

	UA_VALIDATE_INPUT_PARAMETER(user_handle);
	UA_VALIDATE_HANDLE(user_handle, ua_users_list);

	retv_if(NULL == user->account, UA_ERROR_INVALID_PARAMETER);


	g_strlcpy(account, user->account, UAM_USER_ACCOUNT_MAX_STRING_LEN);

	ret = _ua_get_error_code(_uam_remove_user(account));
	if (UA_ERROR_NONE != ret) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		return ret;
		/* LCOV_EXCL_STOP */
	}

	user->isadded = false;

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_get_default_user(ua_user_h* user_handle)
{
	FUNC_ENTRY;
	int ret;
	uam_user_info_s uam_user;
	ua_user_info_s *user = NULL;
	GSList *l;

	UA_VALIDATE_INPUT_PARAMETER(user_handle);

	for (l = ua_users_list; l; l = g_slist_next(l)) {
		user = (ua_user_info_s *)l->data;

		if (user->default_user == true) {
			/* LCOV_EXCL_START */
			UA_INFO("User found [%s]", user->account);
			*user_handle = (ua_user_h)user;
//TODO lk, description says to release handle, but no copy is made of handle. Remove destroy thing from description??
			return UA_ERROR_NONE;
			/* LCOV_EXCL_STOP */
		}
	}


	ret = _ua_get_error_code(_uam_get_default_user(&uam_user));
	if (UA_ERROR_NONE != ret) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		*user_handle = NULL;
		return ret;
		/* LCOV_EXCL_STOP */
	}

	user = g_malloc0(sizeof(ua_user_info_s));
	if (!user) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed to allocated memory");
		return UA_ERROR_OUT_OF_MEMORY;
		/* LCOV_EXCL_STOP */
	}
	user->account = g_strdup(uam_user.account);
	user->name = g_strdup(uam_user.name);
	user->state = UA_PRESENCE_STATE_INACTIVATE;
	user->sensor_bitmask = 0;
	user->user_handle = (ua_user_h)user;

	*user_handle = (ua_user_h)user;
	user->default_user = true;
	user->isadded = true;
	ua_users_list = g_slist_append(ua_users_list, user);
	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_set_account(ua_user_h user_handle, const char *account)
{
	FUNC_ENTRY;
	ua_user_info_s *user = (ua_user_info_s *)user_handle;

	UA_VALIDATE_INPUT_PARAMETER(user_handle);
	UA_VALIDATE_INPUT_PARAMETER(account);
	retv_if((strlen(account) > UAM_USER_ACCOUNT_MAX_STRING_LEN), UA_ERROR_INVALID_PARAMETER);
//TODO lk, check is_added && validate_handle

	if (user->account) {
		/* LCOV_EXCL_START */
		g_free(user->account);
		user->account = NULL;
		/* LCOV_EXCL_STOP */
	}
	user->account = g_strdup(account);

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_get_account(
		ua_user_h user_handle,
		char **account)
{
	FUNC_ENTRY;
	ua_user_info_s *user = (ua_user_info_s *)user_handle;

	UA_VALIDATE_INPUT_PARAMETER(user_handle);
	UA_VALIDATE_INPUT_PARAMETER(account);
	UA_PRINT_USER_HANDLE(user_handle);

	*account = g_strdup(user->account);

	if (*account == NULL) {
		/* LCOV_EXCL_START */
		UA_ERR("g_malloc0 failed");
		return UA_ERROR_OUT_OF_MEMORY;
		/* LCOV_EXCL_STOP */
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_get_name(
		ua_user_h user_handle,
		char **name)
{
	FUNC_ENTRY;
	ua_user_info_s *user = (ua_user_info_s *)user_handle;

	UA_VALIDATE_INPUT_PARAMETER(user_handle);
	UA_VALIDATE_INPUT_PARAMETER(name);
	UA_PRINT_USER_HANDLE(user_handle);

	retv_if(NULL == user->name, UA_ERROR_INVALID_PARAMETER);

	*name = g_strdup(user->name);

	if (*name == NULL) {
		/* LCOV_EXCL_START */
		UA_ERR("g_malloc0 failed");
		return UA_ERROR_OUT_OF_MEMORY;
		/* LCOV_EXCL_STOP */
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_get_last_presence(ua_user_h user_handle, long *timestamp)
{
	FUNC_ENTRY;
	ua_user_info_s *user = (ua_user_info_s *)user_handle;

	UA_VALIDATE_INPUT_PARAMETER(user_handle);
	UA_VALIDATE_INPUT_PARAMETER(timestamp);
	UA_PRINT_USER_HANDLE(user_handle);

	/* update user list from uam_api */

	*timestamp = user->last_presence_timestamp;

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_set_name(
		ua_device_h handle,
		const char *name)
{
	FUNC_ENTRY;
	ua_user_info_s *user = (ua_user_info_s *)handle;

	UA_VALIDATE_INPUT_PARAMETER(handle);
	UA_VALIDATE_INPUT_PARAMETER(name);
	UA_VALIDATE_HANDLE(handle, ua_users_list);

	retv_if(user->isadded, UA_ERROR_INVALID_PARAMETER);
	retv_if((strlen(name) > UAM_USER_NAME_MAX_STRING_LEN), UA_ERROR_INVALID_PARAMETER);

	if (user->name)
		g_free(user->name); // LCOV_EXCL_LINE
	user->name = g_strdup(name);

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_get_handle_by_account(const char* account, ua_user_h *user_handle)
{
	FUNC_ENTRY;

	UA_VALIDATE_INPUT_PARAMETER(account);
	UA_VALIDATE_INPUT_PARAMETER(user_handle);

	*user_handle = _ua_get_user_handle_by_account(account);
	if (*user_handle == NULL) {
		UA_ERR("Failed to get user_handle for given account %s)", account);
		return UA_ERROR_INVALID_PARAMETER;
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_destroy(ua_user_h user_handle)
{
	FUNC_ENTRY;
	ua_user_info_s *user = (ua_user_info_s *)user_handle;
	ua_dev_info_s* dev;
	GSList *l;

	UA_VALIDATE_INPUT_PARAMETER(user_handle);
	UA_VALIDATE_HANDLE(user_handle, ua_users_list);

	retv_if(FALSE == user->create_by_app, UA_ERROR_INVALID_PARAMETER);

	/* Do not remove from list if user is added to database */
	if (user->isadded) {
		/* LCOV_EXCL_START */
		user->create_by_app = FALSE;
		return UA_ERROR_NONE;
		/* LCOV_EXCL_STOP */
	}

	for (l = ua_devices_list; l; l = g_slist_next(l)) {
		dev = (ua_dev_info_s *)l->data;

		if (dev->user == user_handle) {
			/* LCOV_EXCL_START */
			UA_INFO("User found in device handle");
			dev->user = NULL;
			/* LCOV_EXCL_STOP */
		}
	}

	ua_users_list = g_slist_remove(ua_users_list, user);
	_ua_free_ua_user_info_t(user);

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_foreach_added(
		ua_registered_user_cb foreach_cb,
		void *user_data)
{
	FUNC_ENTRY;
	GSList *l;
	ua_user_info_s *u;

	UA_VALIDATE_INPUT_PARAMETER(foreach_cb);

	for (l = ua_users_list; l; l = g_slist_next(l)) {
		u = (ua_user_info_s *)l->data;

		if (!foreach_cb(u->user_handle, user_data))
			break;
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_add_device(ua_user_h user_handle, ua_device_h device_handle,
		ua_user_device_added_cb callback, void *user_data)
{
	FUNC_ENTRY;
	int ret;
	ua_user_h handle = NULL;
	ua_user_h temp_handle = NULL;
	ua_dev_info_s *device = (ua_dev_info_s *) device_handle;
	ua_user_info_s* user = NULL;
	uam_device_info_s uam_device;

	UA_VALIDATE_INPUT_PARAMETER(device_handle);
	UA_VALIDATE_HANDLE(device_handle, ua_devices_list);
	retv_if((((device->device_id == NULL) && (device->mac == NULL))), UA_ERROR_INVALID_PARAMETER);
	retv_if((UA_MAC_TYPE_NONE >= device->type), UA_ERROR_INVALID_PARAMETER);
	retv_if((UA_MAC_TYPE_INVALID <= device->type), UA_ERROR_INVALID_PARAMETER);

	temp_handle = device->user;
	if (user_handle	== NULL) {
		/* LCOV_EXCL_START */
		ret = ua_user_get_default_user(&handle);
		if (ret != UA_ERROR_NONE) {
			UA_ERR("Failed to get default user with error");
			return ret;
		}

		device->user = (ua_user_h)handle;
		/* LCOV_EXCL_STOP */
	} else {
		UA_VALIDATE_HANDLE(user_handle, ua_users_list);
		device->user = user_handle;
	}

	user = (ua_user_info_s *)device->user;
	memset(&uam_device, 0, sizeof(uam_device_info_s));

	uam_device.operating_system = device->os;
	uam_device.type = _ua_to_uam_tech_type(device->type);
	uam_device.discriminant = device->discriminant;

	if (device->mac)
		g_strlcpy(uam_device.mac, device->mac, UAM_MAC_ADDRESS_STRING_LEN);

	g_strlcpy(uam_device.device_id, device->device_id, UA_MOBILE_ID_STRING_LEN);

	if (device->ipv4)
		g_strlcpy(uam_device.ipv4_addr, device->ipv4, UA_IPV4_ADDRESS_STRING_LEN);

	_ua_get_uam_payload_from_ua(&uam_device.payload, device->payload);

	ret = _ua_get_error_code(_uam_request_add_device(user->account, &uam_device));
	if (UA_ERROR_NONE != ret) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		device->user = temp_handle;
		return ret;
		/* LCOV_EXCL_STOP */
	}

	user_callbacks[UA_USER_EVENT_DEVICE_ADDED].callback = callback;
	user_callbacks[UA_USER_EVENT_DEVICE_ADDED].user_data = user_data;
	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_remove_device(ua_user_h user_handle, ua_device_h device_handle)
{
	FUNC_ENTRY;
	int ret;
	ua_user_h handle = NULL;
	ua_dev_info_s *device = (ua_dev_info_s *) device_handle;
	ua_user_info_s* user = NULL;
	uam_device_info_s uam_device;

	UA_VALIDATE_INPUT_PARAMETER(device_handle);
	UA_VALIDATE_HANDLE(device_handle, ua_devices_list);
	retv_if((((device->device_id == NULL) && (device->mac == NULL))), UA_ERROR_INVALID_PARAMETER);
	retv_if((UA_MAC_TYPE_NONE >= device->type), UA_ERROR_INVALID_PARAMETER);
	retv_if((UA_MAC_TYPE_INVALID <= device->type), UA_ERROR_INVALID_PARAMETER);

	if (user_handle == NULL) {
		/* LCOV_EXCL_START */
		ret = ua_user_get_default_user(&handle);
		if (ret != UA_ERROR_NONE) {
			UA_ERR("Failed to get default user with error");
			return ret;
		}

		user = (ua_user_info_s *)handle;
		/* LCOV_EXCL_STOP */
	} else {
		UA_VALIDATE_HANDLE(user_handle, ua_users_list);
		retv_if((device->user != user_handle), UA_ERROR_INVALID_PARAMETER);
		user = (ua_user_info_s *)user_handle;
	}

	memset(&uam_device, 0, sizeof(uam_device_info_s));
	uam_device.operating_system = device->os;
	uam_device.type = _ua_to_uam_tech_type(device->type);

	if (device->mac)
		g_strlcpy(uam_device.mac, device->mac, UAM_MAC_ADDRESS_STRING_LEN);

	g_strlcpy(uam_device.device_id, device->device_id, UA_MOBILE_ID_STRING_LEN);

	if (device->ipv4)
		g_strlcpy(uam_device.ipv4_addr, device->ipv4, UA_IPV4_ADDRESS_STRING_LEN);

//TODO lk, do we need this info to be filled
	_ua_get_uam_payload_from_ua(&uam_device.payload, device->payload);

	ret = _ua_get_error_code(_uam_request_remove_device(user->account, &uam_device));

	if (UA_ERROR_NONE != ret) {
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		return ret;
	}

	device->isadded = false;
	device->user = NULL;
	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_remove_device_by_mac_address(const char *mac)
{
	FUNC_ENTRY;
	int ret;

	UA_VALIDATE_INPUT_PARAMETER(mac);
	ret = _ua_get_error_code(_uam_request_remove_device_by_mac(mac));

	if (UA_ERROR_NONE != ret) {
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		return ret;
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_user_remove_device_by_device_id(const char *device_id, ua_mac_type_e mac_type)
{
	FUNC_ENTRY;
	int ret;

	UA_VALIDATE_INPUT_PARAMETER(device_id);
	retv_if(UA_MAC_TYPE_NONE >= mac_type, UA_ERROR_INVALID_PARAMETER);
	retv_if((UA_MAC_TYPE_INVALID <= mac_type), UA_ERROR_INVALID_PARAMETER);

	ret = _ua_get_error_code(_uam_request_remove_device_by_deviceid(device_id, mac_type));

	if (UA_ERROR_NONE != ret) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		/* LCOV_EXCL_STOP */
		return ret;
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

/* LCOV_EXCL_STOP */
