/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <glib.h>
#include <ua-api.h>

#include <user-awareness.h>
#include <user-awareness-log.h>
#include <user-awareness-private.h>

ua_mac_type_e _to_ua_mac_type(uam_tech_type_e tech_type)
{
	switch (tech_type) {
	case UAM_TECH_TYPE_BT:
		return UA_MAC_TYPE_BT;
	case UAM_TECH_TYPE_BLE:
		return UA_MAC_TYPE_BLE;
	case UAM_TECH_TYPE_WIFI:
		return UA_MAC_TYPE_WIFI;
	case UAM_TECH_TYPE_P2P:
		return UA_MAC_TYPE_P2P;
	default:
		return UA_MAC_TYPE_INVALID;
	}
}

int _ua_get_error_code(int ua_error)
{
	switch (ua_error) {
	case UAM_ERROR_NONE:
		return UA_ERROR_NONE;
	case UAM_ERROR_INVALID_PARAMETER:
		return UA_ERROR_INVALID_PARAMETER;
	/* LCOV_EXCL_START */
	case UAM_ERROR_RESOURCE_BUSY:
		return UA_ERROR_RESOURCE_BUSY;
	case UAM_ERROR_PERMISSION_DENIED:
		return UA_ERROR_PERMISSION_DENIED;
	case UAM_ERROR_TIMED_OUT:
		return UA_ERROR_TIMED_OUT;
	case UAM_ERROR_NOW_IN_PROGRESS:
		return UA_ERROR_NOW_IN_PROGRESS;
	case UAM_ERROR_NOT_SUPPORTED:
		return UA_ERROR_NOT_SUPPORTED;
	case UAM_ERROR_NOT_INITIALIZED:
		return UA_ERROR_NOT_INITIALIZED;
	case UAM_ERROR_NOT_IN_OPERATION:
		return UA_ERROR_NOT_IN_PROGRESS;
	case UAM_ERROR_ALREADY_DONE:
		return UA_ERROR_ALREADY_DONE;
	case UAM_ERROR_INTERNAL:
		return UA_ERROR_OPERATION_FAILED;
	/* LCOV_EXCL_STOP */
	case UAM_ERROR_NOT_FOUND:
		return UA_ERROR_NOT_FOUND;
	/* LCOV_EXCL_START */
	case UAM_ERROR_ALREADY_REGISTERED:
		return UA_ERROR_ALREADY_REGISTERED;
	case UAM_ERROR_DB_FAILED:
		return UA_ERROR_DB_FAILED;
	case UAM_ERROR_NOT_REGISTERED:
		return UA_ERROR_NOT_REGISTERED;
	/* LCOV_EXCL_STOP */
	default:
		return UA_ERROR_OPERATION_FAILED;
	}
}

const char *_ua_get_error_string(ua_error_code_e err)
{
	const char *err_str = NULL;

	switch (err) {
	case UA_ERROR_NONE:
		err_str = "UA_ERROR_NONE";
		break;
	/* LCOV_EXCL_START */
	case UA_ERROR_NOT_PERMITTED:
		err_str = "UA_ERROR_NOT_PERMITTED";
		break;
	/* LCOV_EXCL_STOP */
	case UA_ERROR_INVALID_PARAMETER:
		err_str = "UA_ERROR_INVALID_PARAMETER";
		break;
	/* LCOV_EXCL_START */
	case UA_ERROR_OUT_OF_MEMORY:
		err_str = "UA_ERROR_OUT_OF_MEMORY";
		break;
	case UA_ERROR_PERMISSION_DENIED:
		err_str = "UA_ERROR_PERMISSION_DENIED";
		break;
	case UA_ERROR_RESOURCE_BUSY:
		err_str = "UA_ERROR_RESOURCE_BUSY";
		break;
	case UA_ERROR_TIMED_OUT:
		err_str = "UA_ERROR_TIMED_OUT";
		break;
	case UA_ERROR_CANCELED:
		err_str = "UA_ERROR_CANCELED";
		break;
	case UA_ERROR_NOW_IN_PROGRESS:
		err_str = "UA_ERROR_NOW_IN_PROGRESS";
		break;
	case UA_ERROR_NOT_SUPPORTED:
		err_str = "UA_ERROR_NOT_SUPPORTED";
		break;
	case UA_ERROR_NO_DATA:
		err_str = "UA_ERROR_NO_DATA";
		break;
	case UA_ERROR_NOT_INITIALIZED:
		err_str = "UA_ERROR_NOT_INITIALIZED";
		break;
	case UA_ERROR_NOT_IN_PROGRESS:
		err_str = "UA_ERROR_NOT_IN_PROGRESS";
		break;
	case UA_ERROR_ALREADY_DONE:
		err_str = "UA_ERROR_ALREADY_DONE";
		break;
	/* LCOV_EXCL_STOP */
	case UA_ERROR_OPERATION_FAILED:
		err_str = "UA_ERROR_OPERATION_FAILED";
		break;
	case UA_ERROR_NOT_FOUND:
		err_str = "UA_ERROR_NOT_FOUND";
		break;
	/* LCOV_EXCL_START */
	case UA_ERROR_ALREADY_REGISTERED:
		err_str = "UA_ERROR_ALREADY_REGISTERED";
		break;
	case UA_ERROR_DB_FAILED:
		err_str = "UA_ERROR_DB_FAILED";
		break;
	case UA_ERROR_NOT_REGISTERED:
		err_str = "UA_ERROR_NOT_REGISTERED";
		break;
	default:
		err_str = "NOT Defined";
		break;
	/* LCOV_EXCL_STOP */
	}

	return err_str;
}

uam_tech_type_e _ua_to_uam_tech_type(ua_mac_type_e type)
{
	switch (type) {
	case UA_MAC_TYPE_BT:
		return UAM_TECH_TYPE_BT;
	case UA_MAC_TYPE_BLE:
		return UAM_TECH_TYPE_BLE;
	case UA_MAC_TYPE_WIFI:
		return UAM_TECH_TYPE_WIFI;
	case UA_MAC_TYPE_P2P:
		return UAM_TECH_TYPE_P2P;
	default:
		return UA_MAC_TYPE_INVALID;
	}
}
