/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <gio/gio.h>

#include <user-awareness.h>
#include <user-awareness-internal.h>

#include "uat-menu.h"
#include "uat-common.h"

extern ua_monitor_h g_ua_mon_h;

static char g_sensor_type[MENU_DATA_SIZE + 1] = {0,}; /**< Selected Sensor type */
static char g_sensor_type_str[MENU_DATA_SIZE + 1] = {0,}; /**< Selected Sensor type string */
static char g_presence_threshold[MENU_DATA_SIZE + 1] = {0,}; /**< PRESENCE threashold for litght sensor */
static char g_absence_threshold[MENU_DATA_SIZE + 1] = {0,}; /**< ABSENCE threashold for litght sensor  */

unsigned short g_sensor_list; /**< Added sensor list */
char g_sensor_list_str[MENU_DATA_SIZE + 1] = {0,}; /**< Added sensor list string */

static unsigned int __sensor_type_to_bitmask(int idx)
{
	unsigned int bitmask;
	switch (idx) {
	case UAT_SENSOR_TYPE_BT:
		bitmask = UA_SENSOR_BT;
		break;
	case UAT_SENSOR_TYPE_BLE:
		bitmask = UA_SENSOR_BLE;
		break;
	case UAT_SENSOR_TYPE_WIFI:
		bitmask = UA_SENSOR_WIFI;
		break;
	case UAT_SENSOR_TYPE_MOTION:
		bitmask = UA_SENSOR_MOTION;
		break;
	case UAT_SENSOR_TYPE_LIGHT:
		bitmask = UA_SENSOR_LIGHT;
		break;
	case UAT_SENSOR_TYPE_AUDIO:
		bitmask = UA_SENSOR_AUDIO;
		break;
	default:
		msgr("Unsupported sensor [%d]", idx);
		bitmask = 0x00;
	}

	return bitmask;
}

static bool __ua_monitor_foreach_sensor_cb(ua_sensor_e sensor, void *user_data)
{
	const char *buf = uat_get_sensor_bitmask_str(sensor);
	strncat(g_sensor_list_str, buf,
		sizeof(g_sensor_list_str) - strlen(g_sensor_list_str) - 1);
	strncat(g_sensor_list_str, " ", 2);
	return TRUE;
}

static void __update_sensor_list(void)
{
	memset(g_sensor_list_str, 0, MENU_DATA_SIZE + 1);
	ua_monitor_foreach_sensor(g_ua_mon_h,
				 __ua_monitor_foreach_sensor_cb,
				 NULL);
}

static int run_ua_monitor_add_sensor(
		MManager *mm, struct menu_data *menu)
{
	int sensortype = -1;
	int ret = UA_ERROR_NONE;
	msg("ua_monitor_add_sensor");

	check_if(NULL == g_ua_mon_h);

	if (strlen(g_sensor_type)) {
		sensortype = (unsigned char)strtol(g_sensor_type, NULL, 10);
		snprintf(g_sensor_type_str, MENU_DATA_SIZE, "%s",
			uat_get_sensor_str(sensortype));
	}

	ret = ua_monitor_add_sensor(g_ua_mon_h,
			__sensor_type_to_bitmask(sensortype));

	if (UA_ERROR_NONE == ret)
		__update_sensor_list();

	msg(" - ua_monitor_add_sensor() ret: [0x%X] [%s]",
				ret, uat_get_error_str(ret));

	return RET_SUCCESS;
}

static int run_ua_monitor_remove_sensor(
	MManager *mm, struct menu_data *menu)
{
	int sensortype = -1;
	int ret = UA_ERROR_NONE;

	msg("ua_monitor_remove_sensor");

	check_if(NULL == g_ua_mon_h);

	if (strlen(g_sensor_type)) {
		sensortype = (unsigned char)strtol(g_sensor_type, NULL, 10);
		snprintf(g_sensor_type_str, MENU_DATA_SIZE, "%s",
			uat_get_sensor_str(sensortype));
	}

	ret = ua_monitor_remove_sensor(g_ua_mon_h,
			__sensor_type_to_bitmask(sensortype));

	if (UA_ERROR_NONE == ret)
		__update_sensor_list();

	msg(" - ua_monitor_remove_sensor() ret: [0x%X] [%s]",
				ret, uat_get_error_str(ret));

	return RET_SUCCESS;
}

static int run_ua_monitor_is_sensor_available(
	MManager *mm, struct menu_data *menu)
{
	int sensortype = -1;
	bool is_available = false;
	int ret = UA_ERROR_NONE;

	msg("ua_monitor_is_sensor_available");

	check_if(NULL == g_ua_mon_h);

	if (strlen(g_sensor_type)) {
		sensortype = (unsigned char)strtol(g_sensor_type, NULL, 10);
		snprintf(g_sensor_type_str, MENU_DATA_SIZE, "%s",
			uat_get_sensor_str(sensortype));
	}

	ret = ua_monitor_is_sensor_available(__sensor_type_to_bitmask(sensortype),
			&is_available);

	msg(" - ua_monitor_is_sensor_available() ret: [0x%X] [%s]",
				ret, uat_get_error_str(ret));
	if (UA_ERROR_NONE == ret) {
		if (is_available) {
			msg("%s sensor is available",
				uat_get_sensor_str(sensortype));
		} else {
			msg("%s sensor is not available",
				uat_get_sensor_str(sensortype));
		}
	}

	return RET_SUCCESS;
}

static int run_ua_monitor_foreach_sensor(
	MManager *mm, struct menu_data *menu)
{
	int ret = UA_ERROR_NONE;

	msg("ua_monitor_foreach_sensor");

	check_if(NULL == g_ua_mon_h);

	memset(g_sensor_list_str, 0, MENU_DATA_SIZE + 1);

	ret = ua_monitor_foreach_sensor(g_ua_mon_h, __ua_monitor_foreach_sensor_cb, NULL);

	msg(" - ua_monitor_foreach_sensor() ret: [0x%X] [%s]",
				ret, uat_get_error_str(ret));

	return RET_SUCCESS;
}

static int run_ua_monitor_set_brightness_threshold(
	MManager *mm, struct menu_data *menu)
{
	int ret = UA_ERROR_NONE;
	unsigned int presencethreshold = 0;
	unsigned int absencethreshold = 0;

	msg("ua_monitor_set_brightness_threshold");

	check_if(NULL == g_ua_mon_h);

	if (strlen(g_presence_threshold))
		presencethreshold = (unsigned int)strtol(g_presence_threshold, NULL, 10);

	if (strlen(g_absence_threshold))
		absencethreshold = (unsigned int)strtol(g_absence_threshold, NULL, 10);

	ret = ua_monitor_set_brightness_threshold(g_ua_mon_h,
			presencethreshold, absencethreshold);

	msg(" - ua_monitor_set_brightness_threshold() ret: [0x%X] [%s]",
				ret, uat_get_error_str(ret));

	return RET_SUCCESS;
}

static struct menu_data menu_ua_add_sensor[] = {
	{ "1", "type (1:BT 2:BLE 3:Wi-Fi 4:Motion 5:Light 6:Audio)",
		NULL, NULL, g_sensor_type },
	{ "2", "run", NULL,
		run_ua_monitor_add_sensor, NULL },
	{ NULL, NULL, },
};

static struct menu_data menu_ua_remove_sensor[] = {
	{ "1", "type (1:BT 2:BLE 3:Wi-Fi 4:Motion 5:Light 6:Audio)",
		NULL, NULL, g_sensor_type },
	{ "2", "run", NULL,
		run_ua_monitor_remove_sensor, NULL },
	{ NULL, NULL, },
};

static struct menu_data menu_ua_is_available[] = {
	{ "1", "type (1:BT 2:BLE 3:Wi-Fi 4:Motion 5:Light 6:Audio)",
		NULL, NULL, g_sensor_type },
	{ "2", "run", NULL,
		run_ua_monitor_is_sensor_available, NULL },
	{ NULL, NULL, },
};

static struct menu_data menu_ua_set_brightness_threshold[] = {
	{ "1", "Presence Threshold",
		NULL, NULL, g_presence_threshold},
	{ "2", "Absence Threshold",
		NULL, NULL, g_absence_threshold},
	{ "3", "run", NULL,
		run_ua_monitor_set_brightness_threshold, NULL },
	{ NULL, NULL, },
};

struct menu_data menu_ua_sensors[] = {
	{ "1", "ua_monitor_add_sensor",
		menu_ua_add_sensor, NULL, NULL },
	{ "2", "ua_monitor_remove_sensor",
		menu_ua_remove_sensor, NULL, NULL },
	{ "3", "ua_monitor_is_sensor_available",
		menu_ua_is_available, NULL, NULL },
	{ "4", "ua_monitor_is_sensor_available",
		NULL, run_ua_monitor_foreach_sensor, NULL },
	{ "5", "ua_monitor_set_brightness_threshold",
		menu_ua_set_brightness_threshold, NULL, NULL },
	{ NULL, NULL, },
};
